"""Yahoo Finance helpers module."""

# pylint: disable=unused-argument
from datetime import (
    date as dateType,
    datetime,
)
from pathlib import Path
from typing import Any, Literal, Optional, Union

import pandas as pd
import yfinance as yf
from dateutil.relativedelta import relativedelta
from openbb_core.provider.utils.errors import EmptyDataError
from openbb_yfinance.utils.references import INTERVALS, MONTHS, PERIODS


def get_futures_data() -> pd.DataFrame:
    """Return the dataframe of the futures csv file."""
    return pd.read_csv(Path(__file__).resolve().parent / "futures.csv")


def get_futures_curve(symbol: str, date: Optional[dateType]) -> pd.DataFrame:
    """Get the futures curve for a given symbol.

    Parameters
    ----------
    symbol: str
        Symbol to get futures for
    date: Optional[str]
        Optional historical date to get curve for

    Returns
    -------
    pd.DataFrame
        DataFrame with futures curve
    """
    futures_data = get_futures_data()
    try:
        exchange = futures_data[futures_data["Ticker"] == symbol]["Exchange"].values[0]
    except IndexError:
        return pd.DataFrame({"Last Price": [], "expiration": []})

    today = datetime.today()
    futures_index = []
    futures_curve = []
    historical_curve = []
    i = 0
    empty_count = 0
    # Loop through until we find 12 consecutive empty months
    while empty_count < 12:
        future = today + relativedelta(months=i)
        future_symbol = (
            f"{symbol}{MONTHS[future.month]}{str(future.year)[-2:]}.{exchange}"
        )
        data = yf.download(future_symbol, progress=False, ignore_tz=True, threads=False)

        if data.empty:
            empty_count += 1

        else:
            empty_count = 0
            futures_index.append(future.strftime("%b-%Y"))
            futures_curve.append(data["Adj Close"].values[-1])
            if date is not None:
                historical_curve.append(
                    data["Adj Close"].get(date.strftime("%Y-%m-%d"), None)
                )

        i += 1

    if not futures_index:
        return pd.DataFrame({"date": [], "Last Price": []})

    if historical_curve:
        return pd.DataFrame(
            {"Last Price": historical_curve, "expiration": futures_index}
        )
    return pd.DataFrame({"Last Price": futures_curve, "expiration": futures_index})


# pylint: disable=too-many-arguments,unused-argument
def yf_download(
    symbol: str,
    start_date: Optional[Union[str, dateType]] = None,
    end_date: Optional[Union[str, dateType]] = None,
    interval: INTERVALS = "1d",
    period: PERIODS = "max",
    prepost: bool = False,
    actions: bool = False,
    progress: bool = False,
    ignore_tz: bool = True,
    keepna: bool = False,
    repair: bool = False,
    rounding: bool = False,
    group_by: Literal["ticker", "column"] = "ticker",
    adjusted: bool = False,
    **kwargs: Any,
) -> pd.DataFrame:
    """Get yFinance OHLC data for any ticker and interval available."""
    symbol = symbol.upper()
    _start_date = start_date

    if interval in ["60m", "1h"]:
        period = "2y" if period in ["5y", "10y", "max"] else period
        _start_date = None

    if interval in ["2m", "5m", "15m", "30m", "90m"]:
        _start_date = (datetime.now().date() - relativedelta(days=58)).strftime(
            "%Y-%m-%d"
        )

    if interval == "1m":
        period = "5d"
        _start_date = None

    if adjusted is False:
        kwargs = dict(auto_adjust=False, back_adjust=False)

    try:
        data = yf.download(
            tickers=symbol,
            start=_start_date,
            end=None,
            interval=interval,
            period=period,
            prepost=prepost,
            actions=actions,
            progress=progress,
            ignore_tz=ignore_tz,
            keepna=keepna,
            repair=repair,
            rounding=rounding,
            group_by=group_by,
            threads=False,
            **kwargs,
        )
    except ValueError as exc:
        raise EmptyDataError() from exc

    tickers = symbol.split(",")
    if len(tickers) > 1:
        _data = pd.DataFrame()
        for ticker in tickers:
            temp = data[ticker].copy().dropna(how="all")
            if len(temp) > 0:
                temp.loc[:, "symbol"] = ticker
                temp = temp.reset_index().rename(
                    columns={"Date": "date", "Datetime": "date", "index": "date"}
                )
                _data = pd.concat([_data, temp])
        if not _data.empty:
            index_keys = ["date", "symbol"] if "symbol" in _data.columns else "date"
            _data = _data.set_index(index_keys).sort_index()
            data = _data
    if not data.empty:
        data = data.reset_index()
        data = data.rename(columns={"Date": "date", "Datetime": "date"})
        data["date"] = data["date"].apply(pd.to_datetime)
        data = data[data["Open"] > 0]

        if start_date is not None:
            data = data[data["date"] >= pd.to_datetime(start_date)]
        if (
            end_date is not None
            and start_date is not None
            and pd.to_datetime(end_date) > pd.to_datetime(start_date)
        ):
            data = data[
                data["date"] <= (pd.to_datetime(end_date) + relativedelta(days=1))
            ]

        if period not in [
            "max",
            "1d",
            "5d",
            "1wk",
            "1mo",
            "3mo",
            "6mo",
            "1y",
            "2y",
            "5y",
            "10y",
        ]:
            data["date"] = data["date"].dt.strftime("%Y-%m-%d %H:%M:%S")
        if interval not in ["1m", "2m", "5m", "15m", "30m", "90m", "60m", "1h"]:
            data["date"] = data["date"].dt.strftime("%Y-%m-%d")

        if adjusted is False:
            data = data.drop(columns=["Adj Close"])

        data.columns = data.columns.str.lower().str.replace(" ", "_").to_list()

    return data


def df_transform_numbers(data: pd.DataFrame, columns: list) -> pd.DataFrame:
    """Replace abbreviations of numbers with actual numbers."""
    multipliers = {"M": 1e6, "B": 1e9, "T": 1e12}

    def replace_suffix(x, suffix, multiplier):
        return float(str(x).replace(suffix, "")) * multiplier if suffix in str(x) else x

    for col in columns:
        if col == "% Change":
            data[col] = data[col].astype(str).str.replace("%", "").astype(float)
        else:
            for suffix, multiplier in multipliers.items():
                data[col] = data[col].apply(replace_suffix, args=(suffix, multiplier))

    return data



"""Yahoo Finance Market Indices Model."""

# ruff: noqa: SIM105


from datetime import datetime, timedelta
from typing import Any, Dict, List, Optional

import pandas as pd
from dateutil.relativedelta import relativedelta
from openbb_core.provider.abstract.fetcher import Fetcher
from openbb_core.provider.standard_models.market_indices import (
    MarketIndicesData,
    MarketIndicesQueryParams,
)
from openbb_core.provider.utils.descriptions import QUERY_DESCRIPTIONS
from openbb_core.provider.utils.errors import EmptyDataError
from openbb_yfinance.utils.helpers import yf_download
from openbb_yfinance.utils.references import INDICES, INTERVALS, PERIODS
from pandas import to_datetime
from pydantic import Field


class YFinanceMarketIndicesQueryParams(MarketIndicesQueryParams):
    """YFinance Market Indices Query.

    Source: https://finance.yahoo.com/world-indices
    """

    __json_schema_extra__ = {"symbol": ["multiple_items_allowed"]}

    interval: Optional[INTERVALS] = Field(default="1d", description="Data granularity.")
    period: Optional[PERIODS] = Field(
        default="max", description=QUERY_DESCRIPTIONS.get("period", "")
    )
    prepost: bool = Field(default=True, description="Include Pre and Post market data.")
    rounding: bool = Field(default=True, description="Round prices to two decimals?")


class YFinanceMarketIndicesData(MarketIndicesData):
    """YFinance Market Indices Data."""


class YFinanceMarketIndicesFetcher(
    Fetcher[
        YFinanceMarketIndicesQueryParams,
        List[YFinanceMarketIndicesData],
    ]
):
    """Transform the query, extract and transform the data from the Yahoo Finance endpoints."""

    @staticmethod
    def transform_query(params: Dict[str, Any]) -> YFinanceMarketIndicesQueryParams:
        """Transform the query."""
        transformed_params = params
        now = datetime.now().date()

        if params.get("start_date") is None:
            transformed_params["start_date"] = now - relativedelta(years=1)

        if params.get("end_date") is None:
            transformed_params["end_date"] = now

        tickers = params.get("symbol").lower().split(",")

        new_tickers = []
        for ticker in tickers:
            _ticker = ""
            indices = pd.DataFrame(INDICES).transpose().reset_index()
            indices.columns = ["code", "name", "symbol"]

            if ticker in indices["code"].values:
                _ticker = indices[indices["code"] == ticker]["symbol"].values[0]

            if ticker.title() in indices["name"].values:
                _ticker = indices[indices["name"] == ticker.title()]["symbol"].values[0]

            if "^" + ticker.upper() in indices["symbol"].values:
                _ticker = "^" + ticker.upper()

            if ticker.upper() in indices["symbol"].values:
                _ticker = ticker.upper()

            if _ticker != "":
                new_tickers.append(_ticker)

        transformed_params["symbol"] = ",".join(new_tickers)

        return YFinanceMarketIndicesQueryParams(**params)

    # pylint: disable=unused-argument
    @staticmethod
    def extract_data(
        query: YFinanceMarketIndicesQueryParams,
        credentials: Optional[Dict[str, str]],
        **kwargs: Any,
    ) -> List[dict]:
        """Return the raw data from the Yahoo Finance endpoint."""

        data = yf_download(
            symbol=query.symbol,
            start_date=query.start_date,
            end_date=query.end_date,
            interval=query.interval,
            period=query.period,
            prepost=query.prepost,
            rounding=query.rounding,
        )

        if data.empty:
            raise EmptyDataError()

        days = (
            1
            if query.interval in ["1m", "2m", "5m", "15m", "30m", "60m", "1h", "90m"]
            else 0
        )

        if query.start_date:
            if "date" in data.columns:
                data.set_index("date", inplace=True)
                data.index = to_datetime(data.index)

            data = data[
                (data.index >= to_datetime(query.start_date))
                & (data.index <= to_datetime(query.end_date + timedelta(days=days)))
            ]

        data.reset_index(inplace=True)
        data.rename(columns={"index": "date"}, inplace=True)

        return data.to_dict("records")

    @staticmethod
    def transform_data(
        query: YFinanceMarketIndicesQueryParams,
        data: dict,
        **kwargs: Any,
    ) -> List[YFinanceMarketIndicesData]:
        """Transform the data to the standard format."""
        return [YFinanceMarketIndicesData.model_validate(d) for d in data]



"""YFinance Share Statistics Model."""

# pylint: disable=unused-argument
import asyncio
import warnings
from datetime import (
    date as dateType,
    datetime,
)
from typing import Any, Dict, List, Optional

from openbb_core.provider.abstract.fetcher import Fetcher
from openbb_core.provider.standard_models.share_statistics import (
    ShareStatisticsData,
    ShareStatisticsQueryParams,
)
from pydantic import Field, field_validator
from yfinance import Ticker

_warn = warnings.warn


class YFinanceShareStatisticsQueryParams(ShareStatisticsQueryParams):
    """YFinance Share Statistics Query."""

    __json_schema_extra__ = {"symbol": ["multiple_items_allowed"]}


class YFinanceShareStatisticsData(ShareStatisticsData):
    """YFinance Share Statistics Data."""

    __alias_dict__ = {
        "outstanding_shares": "sharesOutstanding",
        "float_shares": "floatShares",
        "date": "dateShortInterest",
    }

    implied_shares_outstanding: Optional[int] = Field(
        default=None,
        description="Implied Shares Outstanding of common equity,"
        + " assuming the conversion of all convertible subsidiary equity into common.",
        alias="impliedSharesOutstanding",
    )
    short_interest: Optional[int] = Field(
        default=None,
        description="Number of shares that are reported short.",
        alias="sharesShort",
    )
    short_percent_of_float: Optional[float] = Field(
        default=None,
        description="Percentage of shares that are reported short, as a normalized percent.",
        json_schema_extra={"unit_measurement": "percent", "frontend_multiply": 100},
        alias="shortPercentOfFloat",
    )
    days_to_cover: Optional[float] = Field(
        default=None,
        description="Number of days to repurchase the shares as a ratio of average daily volume",
        alias="shortRatio",
    )
    short_interest_prev_month: Optional[int] = Field(
        default=None,
        description="Number of shares that were reported short in the previous month.",
        alias="sharesShortPriorMonth",
    )
    short_interest_prev_date: Optional[dateType] = Field(
        default=None,
        description="Date of the previous month's report.",
        alias="sharesShortPreviousMonthDate",
    )
    insider_ownership: Optional[float] = Field(
        default=None,
        description="Percentage of shares held by insiders, as a normalized percent.",
        json_schema_extra={"unit_measurement": "percent", "frontend_multiply": 100},
        alias="heldPercentInsiders",
    )
    institution_ownership: Optional[float] = Field(
        default=None,
        description="Percentage of shares held by institutions, as a normalized percent.",
        json_schema_extra={"unit_measurement": "percent", "frontend_multiply": 100},
        alias="heldPercentInstitutions",
    )
    institution_float_ownership: Optional[float] = Field(
        default=None,
        description="Percentage of float held by institutions, as a normalized percent.",
        json_schema_extra={"unit_measurement": "percent", "frontend_multiply": 100},
        alias="institutionsFloatPercentHeld",
    )
    institutions_count: Optional[int] = Field(
        default=None,
        description="Number of institutions holding shares.",
        alias="institutionsCount",
    )

    @field_validator(
        "date", "short_interest_prev_date", mode="before", check_fields=False
    )
    @classmethod
    def validate_first_trade_date(cls, v):
        """Convert  dates from UTC timestamp."""
        return datetime.utcfromtimestamp(v).date() if v else None


class YFinanceShareStatisticsFetcher(
    Fetcher[YFinanceShareStatisticsQueryParams, List[YFinanceShareStatisticsData]]
):
    """YFinance Share Statistics Fetcher."""

    @staticmethod
    def transform_query(params: Dict[str, Any]) -> YFinanceShareStatisticsQueryParams:
        """Transform the query."""
        return YFinanceShareStatisticsQueryParams(**params)

    @staticmethod
    async def aextract_data(
        query: YFinanceShareStatisticsQueryParams,
        credentials: Optional[Dict[str, str]],
        **kwargs: Any,
    ) -> List[Dict]:
        """Extract the raw data from YFinance."""
        symbols = query.symbol.split(",")
        results = []
        fields = [
            "symbol",
            "sharesOutstanding",
            "floatShares",
            "impliedSharesOutstanding",
            "sharesShort",
            "sharesShortPriorMonth",
            "sharesShortPreviousMonthDate",
            "shortRatio",
            "shortPercentOfFloat",
            "dateShortInterest",
            "heldPercentInsiders",
            "heldPercentInstitutions",
            "institutionsFloatPercentHeld",
            "institutionsCount",
        ]

        async def get_one(symbol):
            """Get the data for one ticker symbol."""
            result = {}
            ticker = {}
            try:
                _ticker = Ticker(symbol)
                ticker = _ticker.get_info()
                major_holders = _ticker.get_major_holders(as_dict=True).get("Value")
                if major_holders:
                    ticker.update(major_holders)  # type: ignore
            except Exception as e:
                _warn(f"Error getting data for {symbol}: {e}")
            if ticker:
                for field in fields:
                    if field in ticker:
                        result[field] = ticker.get(field, None)
                if result and result.get("sharesOutstanding") is not None:
                    results.append(result)

        tasks = [get_one(symbol) for symbol in symbols]

        await asyncio.gather(*tasks)

        return results

    @staticmethod
    def transform_data(
        query: YFinanceShareStatisticsQueryParams,
        data: List[Dict],
        **kwargs: Any,
    ) -> List[YFinanceShareStatisticsData]:
        """Transform the data."""
        return [YFinanceShareStatisticsData.model_validate(d) for d in data]



"""Yahoo Finance Equity Historical Price Model."""

# pylint: disable=unused-argument
# ruff: noqa: SIM105

from datetime import datetime
from typing import Any, Dict, List, Literal, Optional
from warnings import warn

from dateutil.relativedelta import relativedelta
from openbb_core.provider.abstract.fetcher import Fetcher
from openbb_core.provider.standard_models.equity_historical import (
    EquityHistoricalData,
    EquityHistoricalQueryParams,
)
from openbb_core.provider.utils.descriptions import QUERY_DESCRIPTIONS
from openbb_core.provider.utils.errors import EmptyDataError
from openbb_yfinance.utils.helpers import yf_download
from openbb_yfinance.utils.references import INTERVALS_DICT, PERIODS
from pandas import DataFrame, Timestamp
from pydantic import Field, PrivateAttr, field_validator, model_validator


class YFinanceEquityHistoricalQueryParams(EquityHistoricalQueryParams):
    """Yahoo Finance Equity Historical Price Query.

    Source: https://finance.yahoo.com/
    """

    __json_schema_extra__ = {"symbol": ["multiple_items_allowed"]}

    interval: Literal[
        "1m",
        "2m",
        "5m",
        "15m",
        "30m",
        "60m",
        "90m",
        "1h",
        "1d",
        "5d",
        "1W",
        "1M",
        "1Q",
    ] = Field(
        default="1d",
        description=QUERY_DESCRIPTIONS.get("interval", ""),
    )
    extended_hours: bool = Field(
        default=False,
        description="Include Pre and Post market data.",
    )
    include_actions: bool = Field(
        default=True,
        description="Include dividends and stock splits in results.",
    )
    adjustment: Literal["splits_only", "splits_and_dividends"] = Field(
        default="splits_only",
        description="The adjustment factor to apply. Default is splits only.",
    )
    adjusted: bool = Field(
        default=False,
        exclude=True,
        description="This field is deprecated (4.1.5) and will be removed in a future version."
        + " Use 'adjustment' set as 'splits_and_dividends' instead.",
        json_schema_extra={"deprecated": True},
    )
    prepost: bool = Field(
        default=False,
        exclude=True,
        description="This field is deprecated (4.1.5) and will be removed in a future version."
        + " Use 'extended_hours' as True instead.",
        json_schema_extra={"deprecated": True},
    )

    _ignore_tz: bool = PrivateAttr(default=True)
    _progress: bool = PrivateAttr(default=False)
    _keepna: bool = PrivateAttr(default=False)
    _period: PERIODS = PrivateAttr(default="max")
    _rounding: bool = PrivateAttr(default=False)
    _repair: bool = PrivateAttr(default=False)
    _group_by: Literal["ticker", "column"] = PrivateAttr(default="ticker")

    @model_validator(mode="before")
    @classmethod
    def validate_deprecated_params(cls, values):
        """Validate the deprecated parameters."""
        for k, v in values.copy().items():
            if k in ["adjusted"] and v is True:
                warn(
                    f"The '{k}' parameter is deprecated and will be removed in a future version."
                )
                values["adjustment"] = "splits_and_dividends"
            if k in ["prepost"] and v is True:
                warn(
                    f"The '{k}' parameter is deprecated and will be removed in a future version."
                )
                values["extended_hours"] = True
        return values


class YFinanceEquityHistoricalData(EquityHistoricalData):
    """Yahoo Finance Equity Historical Price Data."""

    __alias_dict__ = {
        "split_ratio": "stock_splits",
        "dividend": "dividends",
    }

    split_ratio: Optional[float] = Field(
        default=None,
        description="Ratio of the equity split, if a split occurred.",
    )
    dividend: Optional[float] = Field(
        default=None,
        description="Dividend amount (split-adjusted), if a dividend was paid.",
    )

    @field_validator("date", mode="before", check_fields=False)
    def date_validate(cls, v):  # pylint: disable=E0213
        """Return formatted datetime."""
        if isinstance(v, Timestamp):
            return v.to_pydatetime()
        return v


class YFinanceEquityHistoricalFetcher(
    Fetcher[
        YFinanceEquityHistoricalQueryParams,
        List[YFinanceEquityHistoricalData],
    ]
):
    """Transform the query, extract and transform the data from the Yahoo Finance endpoints."""

    @staticmethod
    def transform_query(params: Dict[str, Any]) -> YFinanceEquityHistoricalQueryParams:
        """Transform the query."""
        transformed_params = params
        now = datetime.now().date()

        if params.get("start_date") is None:
            transformed_params["start_date"] = now - relativedelta(years=1)

        if params.get("end_date") is None:
            transformed_params["end_date"] = now

        return YFinanceEquityHistoricalQueryParams(**params)

    @staticmethod
    def extract_data(
        query: YFinanceEquityHistoricalQueryParams,
        credentials: Optional[Dict[str, str]],
        **kwargs: Any,
    ) -> DataFrame:
        """Return the raw data from the Yahoo Finance endpoint."""
        adjusted = query.adjustment == "splits_and_dividends"
        kwargs = {"auto_adjust": True, "back_adjust": True} if adjusted is True else {}
        # pylint: disable=protected-access
        data = yf_download(
            symbol=query.symbol,
            start_date=query.start_date,
            end_date=query.end_date,
            interval=INTERVALS_DICT[query.interval],  # type: ignore
            period=query._period,
            prepost=query.extended_hours,
            actions=query.include_actions,
            progress=query._progress,
            ignore_tz=query._ignore_tz,
            keepna=query._keepna,
            repair=query._repair,
            rounding=query._rounding,
            group_by=query._group_by,
            adjusted=adjusted,
            **kwargs,
        )

        if data.empty:
            raise EmptyDataError()

        return data

    @staticmethod
    def transform_data(
        query: YFinanceEquityHistoricalQueryParams,
        data: DataFrame,
        **kwargs: Any,
    ) -> List[YFinanceEquityHistoricalData]:
        """Transform the data to the standard format."""
        if "capital_gains" in data.columns:
            data = (
                data.drop(columns=["capital_gains"])
                if query.include_actions is False
                else data
            )
        return [
            YFinanceEquityHistoricalData.model_validate(d)
            for d in data.to_dict("records")
        ]
